@@ -28,8 +28,6 @@
 #' \itemize{
 #'   \item \code{hjust}
 #'   \item \code{vjust}
-#'   \item \code{nudge_x}
-#'   \item \code{nudge_y}
 #'   \item \code{position}
 #'   \item \code{check_overlap}
 #' }
@@ -58,12 +56,14 @@
 #'   three types of arguments you can use here:
 #'
 #'   \itemize{
-#'   \item Aesthetics: to set an aesthetic to a fixed value, like
-#'      \code{color = "red"} or \code{size = 3}.
-#'   \item Other arguments to the layer, for example you override the
-#'     default \code{stat} associated with the layer.
-#'   \item Other arguments passed on to the stat.
+#'     \item Aesthetics: to set an aesthetic to a fixed value, like
+#'        \code{color = "red"} or \code{size = 3}.
+#'     \item Other arguments to the layer, for example you override the
+#'       default \code{stat} associated with the layer.
+#'     \item Other arguments passed on to the stat.
 #'   }
+#' @param nudge_x,nudge_y Horizontal and vertical adjustments to nudge the
+#'   starting position of each text label.
 #' @param box.padding Amount of padding around bounding box. Defaults to
 #'   \code{unit(0.25, "lines")}.
 #' @param point.padding Amount of padding around labeled point. Defaults to
@@ -71,6 +71,7 @@
 #' @param segment.color Color of the line segment connecting the data point to
 #'   the text labe. Defaults to \code{#666666}.
 #' @param segment.size Width of segment, in mm.
+#' @param arrow specification for arrow heads, as created by \code{\link[grid]{arrow}}
 #' @param force Force of repulsion between overlapping text labels. Defaults
 #'   to 1.
 #' @param max.iter Maximum number of iterations to try to resolve overlaps.
@@ -93,6 +94,10 @@
 #' p + geom_text_repel(aes(colour = factor(cyl)))
 #' p + geom_label_repel(aes(fill = factor(cyl)), colour = "white", fontface = "bold")
 #'
+#' # Nudge the starting positions
+#' p + geom_text_repel(nudge_x = ifelse(mtcars$cyl == 6, 1, 0),
+#'                     nudge_y = ifelse(mtcars$cyl == 6, 8, 0))
+#' # Change the text size
 #' p + geom_text_repel(aes(size = wt))
 #' # Scale height of text, rather than sqrt(height)
 #' p + geom_text_repel(aes(size = wt)) + scale_radius(range = c(3,6))
@@ -107,19 +112,26 @@
 #' p +
 #'   geom_text_repel() +
 #'   annotate("text", label = "plot mpg vs. wt", x = 2, y = 15, size = 8, colour = "red")
-#' }
 #'
+#' # Add arrows
+#' p +
+#'   geom_point(colour = "red") +
+#'   geom_text_repel(arrow = arrow(length = unit(0.02, "npc")), box.padding = unit(1, "lines"))
+#' }
 #' @export
 geom_text_repel <- function(
   mapping = NULL, data = NULL, stat = "identity",
   parse = FALSE,
   ...,
   box.padding = unit(0.25, "lines"),
-  point.padding = unit(0, "lines"),
+  point.padding = unit(1e-6, "lines"),
   segment.color = "#666666",
   segment.size = 0.5,
+  arrow = NULL,
   force = 1,
   max.iter = 2000,
+  nudge_x = 0,
+  nudge_y = 0,
   na.rm = FALSE,
   show.legend = NA,
   inherit.aes = TRUE
@@ -139,8 +151,11 @@ geom_text_repel <- function(
       point.padding = point.padding,
       segment.color = segment.color,
       segment.size = segment.size,
+      arrow = arrow,
       force = force,
       max.iter = max.iter,
+      nudge_x = nudge_x,
+      nudge_y = nudge_y,
       ...
     )
   )
@@ -164,11 +179,14 @@ GeomTextRepel <- ggproto("GeomTextRepel", Geom,
     parse = FALSE,
     na.rm = FALSE,
     box.padding = unit(0.25, "lines"),
-    point.padding = unit(0, "lines"),
+    point.padding = unit(1e-6, "lines"),
     segment.color = "#666666",
     segment.size = 0.5,
+    arrow = NULL,
     force = 1,
-    max.iter = 2000
+    max.iter = 2000,
+    nudge_x = 0,
+    nudge_y = 0
   ) {
     lab <- data$label
     if (parse) {
@@ -179,88 +197,119 @@ GeomTextRepel <- ggproto("GeomTextRepel", Geom,
     limits <- data.frame(x = panel_scales$x.range, y = panel_scales$y.range)
     limits <- coord$transform(limits, panel_scales)
 
+    # Transform the nudges to the panel scales.
+    nudges <- data.frame(
+      x = data$x + nudge_x, y = data$y + nudge_y
+      # x = rep_len(nudge_x, nrow(data)),
+      # y = rep_len(nudge_y, nrow(data))
+    )
+    nudges <- coord$transform(nudges, panel_scales)
+
     # Transform the raw data to the panel scales.
     data <- coord$transform(data, panel_scales)
 
-    # The padding around each bounding box.
-    pad.x <- convertWidth(box.padding, "npc", valueOnly = TRUE)
-    pad.y <- convertHeight(box.padding, "npc", valueOnly = TRUE)
+    # The nudge is relative to the data.
+    nudges$x <- nudges$x - data$x
+    nudges$y <- nudges$y - data$y
 
-    # Create a dataframe with x1 y1 x2 y2
-    boxes <- lapply(1:nrow(data), function(i) {
-      row <- data[i, , drop = FALSE]
-      tg <- textGrob(
-        lab[i],
-        row$x, row$y, default.units = "native",
-        rot = row$angle,
-        gp = gpar(
-          col = alpha(row$colour, row$alpha),
-          fontsize = row$size * .pt,
-          fontfamily = row$family,
-          fontface = row$fontface,
-          lineheight = row$lineheight
-        ),
-        check.overlap = FALSE
-      )
-      c(
-        "x1" = row$x + convertWidth(grobX(tg, "west"), "npc", TRUE) - pad.x,
-        "y1" = row$y - convertHeight(grobHeight(tg), "npc", TRUE) / 2 - pad.y,
-        "x2" = row$x + convertWidth(grobX(tg, "east"), "npc", TRUE) + pad.x,
-        "y2" = row$y + convertHeight(grobHeight(tg), "npc", TRUE) / 2 + pad.y
-      )
-    })
-
-    # Fudge factor to make each box slightly wider. This is useful when the
-    # user adds a legend to the plot, causing all the labels to squeeze
-    # together.
-    fudge.width <- abs(max(limits$x) - min(limits$x)) / 80
-    boxes <- lapply(boxes, function(b) {
-      # fudge.width <- abs(b['x2'] - b['x1']) / 10
-      b['x1'] <- b['x1'] - fudge.width
-      b['x2'] <- b['x2'] + fudge.width
-      b
-    })
-
-    # Repel overlapping bounding boxes away from each other.
-    repel <- repel_boxes(
-      do.call(rbind, boxes),
-      xlim = range(limits$x),
-      ylim = range(limits$y),
-      force = force * 1e-6,
-      maxiter = max.iter
-    )
-
-    grobs <- lapply(1:nrow(data), function(i) {
-      row <- data[i, , drop = FALSE]
-      textRepelGrob(
-        lab[i],
-        x = unit(repel$x[i], "native"),
-        y = unit(repel$y[i], "native"),
-        x.orig = unit(data$x[i], "native"),
-        y.orig = unit(data$y[i], "native"),
-        box.padding = box.padding,
-        point.padding = point.padding,
-        text.gp = gpar(
-          col = row$colour,
-          fontsize = row$size * .pt,
-          fontfamily = row$family,
-          fontface = row$fontface,
-          lineheight = row$lineheight
-        ),
-        segment.gp = gpar(
-          col = segment.color,
-          lwd = segment.size * .pt
-        )
-      )
-    })
-    class(grobs) <- "gList"
-
-    ggname("geom_text_repel", grobTree(children = grobs))
+    ggname("geom_text_repel", gTree(
+      limits = limits,
+      data = data,
+      lab = lab,
+      nudges = nudges,
+      box.padding = box.padding,
+      point.padding = point.padding,
+      segment.color = segment.color,
+      segment.size = segment.size,
+      arrow = arrow,
+      force = force,
+      max.iter = max.iter,
+      cl = "textrepeltree"
+    ))
   },
 
   draw_key = draw_key_text
 )
 
+#' grid::makeContent function for the grobTree of textRepelGrob objects
+#' @param x A grid grobTree.
+#' @export
+#' @noRd
+makeContent.textrepeltree <- function(x) {
+
+  # The padding around each bounding box.
+  pad.x <- convertWidth(x$box.padding, "native", valueOnly = TRUE)
+  pad.y <- convertHeight(x$box.padding, "native", valueOnly = TRUE)
+
+  # The padding around each point.
+  pad.point.x <- convertWidth(x$point.padding, "native", valueOnly = TRUE)
+  pad.point.y <- convertHeight(x$point.padding, "native", valueOnly = TRUE)
+
+  # Create a dataframe with x1 y1 x2 y2
+  boxes <- lapply(1:nrow(x$data), function(i) {
+    row <- x$data[i, , drop = FALSE]
+    tg <- textGrob(
+      x$lab[i],
+      row$x, row$y, default.units = "native",
+      rot = row$angle,
+      gp = gpar(
+        fontsize = row$size * .pt,
+        fontfamily = row$family,
+        fontface = row$fontface,
+        lineheight = row$lineheight
+      )
+    )
+    gw <- convertWidth(grobWidth(tg), "native", TRUE) / 2
+    gh <- convertHeight(grobHeight(tg), "native", TRUE) / 2
+    c(
+      "x1" = row$x - gw - pad.x + x$nudges$x[i],
+      "y1" = row$y - gh - pad.y + x$nudges$y[i],
+      "x2" = row$x + gw + pad.x + x$nudges$x[i],
+      "y2" = row$y + gh + pad.y + x$nudges$y[i]
+    )
+  })
+
+  # Repel overlapping bounding boxes away from each other.
+  repel <- repel_boxes(
+    data_points = cbind(x$data$x, x$data$y),
+    pad_point_x = pad.point.x,
+    pad_point_y = pad.point.y,
+    boxes = do.call(rbind, boxes),
+    xlim = range(x$limits$x),
+    ylim = range(x$limits$y),
+    force = x$force * 1e-6,
+    maxiter = x$max.iter
+  )
+
+  grobs <- lapply(1:nrow(x$data), function(i) {
+    row <- x$data[i, , drop = FALSE]
+    textRepelGrob(
+      x$lab[i],
+      x = unit(repel$x[i], "native"),
+      y = unit(repel$y[i], "native"),
+      x.orig = unit(x$data$x[i], "native"),
+      y.orig = unit(x$data$y[i], "native"),
+      box.padding = x$box.padding,
+      point.padding = x$point.padding,
+      text.gp = gpar(
+        col = row$colour,
+        fontsize = row$size * .pt,
+        fontfamily = row$family,
+        fontface = row$fontface,
+        lineheight = row$lineheight
+      ),
+      segment.gp = gpar(
+        col = x$segment.color,
+        lwd = x$segment.size * .pt
+      ),
+      arrow = x$arrow
+    )
+  })
+  class(grobs) <- "gList"
+
+  setChildren(x, grobs)
+}
+
 textRepelGrob <- function(
   label,
   x = unit(0.5, "npc"),
@@ -270,11 +319,12 @@ textRepelGrob <- function(
   default.units = "npc",
   just = "center",
   box.padding = unit(0.25, "lines"),
-  point.padding = unit(0, "lines"),
+  point.padding = unit(1e-6, "lines"),
   name = NULL,
   text.gp = gpar(),
   segment.gp = gpar(),
-  vp = NULL
+  vp = NULL,
+  arrow = NULL
 ) {
 
   stopifnot(length(label) == 1)
@@ -297,7 +347,8 @@ textRepelGrob <- function(
     text.gp = text.gp,
     segment.gp = segment.gp,
     vp = vp,
-    cl = "textrepelgrob"
+    cl = "textrepelgrob",
+    arrow = arrow
   )
 }
 
@@ -305,6 +356,7 @@ textRepelGrob <- function(
 #'
 #' @param x A grid grob.
 #' @export
+#' @noRd
 makeContent.textrepelgrob <- function(x) {
   hj <- resolveHJust(x$just, NULL)
   vj <- resolveVJust(x$just, NULL)
@@ -330,32 +382,29 @@ makeContent.textrepelgrob <- function(x) {
 
   center <- centroid(c(x1, y1, x2, y2))
 
-  d <- (center - orig)
-  d <- d / euclid(center, orig)
-  orig <- orig + convertWidth(x$point.padding, "native", TRUE) * d
-
-  pad.x <- convertWidth(x$box.padding, "native", TRUE) / 2
-  pad.y <- convertHeight(x$box.padding, "native", TRUE) / 2
+  # Get the coordinates of the intersection between the line from the
+  # original data point to the centroid and the rectangle's edges.
+  pad.x <- convertWidth(unit(0.25, "lines"), "native", TRUE) / 2
+  pad.y <- convertHeight(unit(0.25, "lines"), "native", TRUE) / 2
   b <- c(x1 - pad.x, y1 - pad.y, x2 + pad.x, y2 + pad.y)
+  int <- intersect_line_rectangle(orig, center, b)
 
-  if (!point_within_box(orig, b)) {
+  # Nudge the original data point toward the label with point.padding.
+  pad.x <- convertWidth(x$point.padding, "native", TRUE) / 2
+  pad.y <- convertHeight(x$point.padding, "native", TRUE) / 2
+  b <- c(orig[1] - pad.x, orig[2] - pad.y, orig[1] + pad.x, orig[2] + pad.y)
+  orig <- intersect_line_rectangle(center, orig, b)
 
-    # Get the coordinates of the intersection between the line from the
-    # original data point to the centroid and the rectangle's edges.
-    int <- intersect_line_rectangle(orig, center, b)
+  s <- segmentsGrob(
+    x0 = int[1],
+    y0 = int[2],
+    x1 = orig[1],
+    y1 = orig[2],
+    default.units = "native",
+    gp = x$segment.gp,
+    name = "segment",
+    arrow = x$arrow
+  )
 
-    s <- segmentsGrob(
-      x0 = int[1],
-      y0 = int[2],
-      x1 = orig[1],
-      y1 = orig[2],
-      default.units = "native",
-      gp = x$segment.gp,
-      name = "segment"
-    )
-
-    return(setChildren(x, gList(s, t)))
-  }
-
-  return(setChildren(x, gList(t)))
+  setChildren(x, gList(s, t))
 }
